﻿import { mfunction } from './../mfunction'
import expressionData from '../../expressionData'
import { expressionPart } from './../../compiler/parts/expressionPart'
import { List } from 'typescript-dotnet-commonjs/System/Collections/List'
import { groupDataContext } from './../../functions/groupDataContext';
import { KeyValuePair } from 'typescript-dotnet-commonjs/System/KeyValuePair'
import { IEnumerable } from 'typescript-dotnet-commonjs/System/Collections/Enumeration/IEnumerable';
import { literalPart } from '../../compiler/parts/literalPart';
import { DateTime } from 'typescript-dotnet-commonjs/System/Time/DateTime'
import { compileFlags } from '../../compiler/expressionCompiler';
import { getType } from '../../utils/reflectionExtender';
import { typeMapper } from '../../types/typeMapper';
import { namedSelection } from '../namedSelection';
import { outputVariables } from '../../utils/outputVariables';
import { utils } from '../../compiler/utils';
import { isFloatType, isNumberType } from '../../utils/general';
import ICollection from 'typescript-dotnet-commonjs/System/Collections/ICollection';
import { selection } from '../selection';
import { DictionaryObject } from '../../utils/dictionaryObject';

export class toList extends mfunction {

    public get minParameters(): number {
        return 1;
    }

    public get maxParameters(): number {
        return -1;
    }

    public compileParameters(paramList: List<string>, flags: compileFlags): List<expressionPart> {
        if (flags.legacyMode) {
            let parts: List<expressionPart> = new List<expressionPart>();
            for (let param of paramList.toArray()) {
                let lPart = new literalPart();
                lPart.value = param.trim();
                parts.add(lPart);
            }
            return parts;
        }
        else {
            return super.compileParameters(paramList, flags);
        }

    }

    public invoke(expressionData: expressionData, parameters: List<expressionPart>, /* out */resolved: outputVariables, running: groupDataContext | undefined): Object {
        resolved.resolved = true;
        let result: List<Object> = new List<Object>();
        for (let ex of parameters.toArray()) {
            let str: Object = ex.resolve(expressionData, resolved, running) as Object;
            if(str !== undefined)
                result.add(str);
        }

        resolved.resolved = true;
        return result;
    }
}

export class toDictionary extends mfunction {

    public /* override */ get minParameters(): number {
        return 1;
    }

    public /* override */ get maxParameters(): number {
        return -1;
    }

    public invoke(expressionData: expressionData, parameters: List<expressionPart>, /* out */resolved: outputVariables, running?: groupDataContext): Object {
        resolved.resolved = true;
        let result: DictionaryObject = new DictionaryObject();
        for (let ex of parameters.toArray()) {
            let object;
            let val = (ex.resolveGeneric<DictionaryObject>(
                expressionData,
                resolved,
                getType(new DictionaryObject()),
                running));
            let key = val.getKeys()[0];
            let valueOfDic = val.getValue(key);
            result.add(key, valueOfDic as Object );
        }

        resolved.resolved = true;
        return result;
    }
}

export class toDate extends mfunction {

    public /* override */ get minParameters(): number {
        return 1;
    }

    public /* override */ get maxParameters(): number {
        return 1;
    }

    public invoke(expressionData: expressionData, parameters: List<expressionPart>, /* out */resolved: outputVariables, running?: groupDataContext): Object | undefined {
        let value: unknown = null;
        resolved.resolved = true;
        if ((parameters.count > 0)) {
            value = parameters.get(0).resolve(expressionData, /* out */resolved, /* running:*/ undefined);
        }

        let paramDate: DateTime;
        if ((resolved
            && (value != null))) {
            let paramIsDate: boolean = false;
            if ((value instanceof DateTime)) {
                paramDate = (<DateTime>(value));
                paramIsDate = true;
            }
            else {
                //paramIsDate =  DateTime.TryParse(value.toString(), /* out */paramDate);
                paramDate = new DateTime((value as string).toString());
                paramIsDate = true;
           }

            if (paramIsDate) {
                //Object.values[]
                return Date.parse(Object.values(paramDate)[1]);
            }

        }

        return undefined;
    }
}

export class toNumber extends mfunction {

    public /* override */ get minParameters(): number {
        return 1;
    }

    public /* override */ get maxParameters(): number {
        return 2;
    }

    public invoke(expressionData: expressionData, parameters: List<expressionPart>, /* out */resolved: outputVariables, running: groupDataContext | undefined): Object | undefined {
        let value: Object | undefined = undefined;
        resolved.resolved = true;
        let type: string = "auto";
        if ((parameters.count > 0)) {
            value = parameters.get(0).resolve(expressionData, /* out */resolved, /* running:*/ undefined);
        }

        if ((parameters.count > 1)) {
            type = parameters.get(1).resolve(expressionData, /* out */resolved, /* running:*/ undefined) as string;
        }

        if ((resolved.resolved
            && (value != null))) {
            let str: string = value.toString();
            switch (type) {
                // case "sbyte":
                //     return typeMapper.toType<System.SByte>(value);
                //     break;
                case "short":
                    return typeMapper.toType<number>(value);
                    break;
                case "int":
                    return typeMapper.toType<number>(value);
                    break;
                case "long":
                    return typeMapper.toType<number>(value);
                    break;
                case "byte":
                    return typeMapper.toType<number>(value);
                    break;
                case "ushort":
                    return typeMapper.toType<number>(value);
                    break;
                case "uint":
                    return typeMapper.toType<number>(value);
                    break;
                case "ulong":
                    return typeMapper.toType<number>(value);
                    break;
                case "float":
                    return typeMapper.toType<number>(value);
                    break;
                case "double":
                    return typeMapper.toType<number>(value);
                    break;
                case "auto":
                default:
                    let tempInt: number;
                    let d = value as number;
                    if (isNumberType(value) && isFloatType(value as number)) {
                        return Number.parseFloat(value.toString());
                    }
                    else if (isNumberType(value)) {
                        return Number.parseInt(value.toString());
                    }
                    else
                        return value;
                    // if (number.TryParse((value + ""), /* out */tempInt)) {
                    //     return tempInt;
                    // }
                    // else {
                    //     return typeMapper.toType<number>(value);
                    // }

                    break;
            }

        }
        else {
            switch (type) {
                case "short":
                    return (<number>(0));
                    break;
                case "int":
                    return (<number>(0));
                    break;
                case "long":
                    return (<number>(0));
                    break;
                case "byte":
                    return (<number>(0));
                    break;
                case "ushort":
                    return (<number>(0));
                    break;
                case "uint":
                    return (<number>(0));
                    break;
                case "ulong":
                    return (<number>(0));
                    break;
                case "float":
                    return (<number>(0));
                    break;
                case "double":
                    return (<number>(0));
                    break;
                case "auto":
                    break;
                default:
                    return 0;
                    break;
            }

        }

    }
}

export class coalesce extends mfunction {

    public get minParameters(): number {
        return 2;
    }

    public get maxParameters(): number {
        return 3;
    }

    public invoke(expressionData: expressionData, parameters: List<expressionPart>, /* out */resolved: outputVariables, running: groupDataContext | undefined): Object | undefined {
        resolved.resolved = true;
        if ((parameters.count > 1)) {
            let subResolved: outputVariables = new outputVariables();
            let value: Object | undefined = parameters.get(0).resolve(expressionData, /* out */subResolved, /* running:*/ undefined);
            let emptyAsNull = parameters.count > 2 ? parameters.get(2).resolveGeneric<boolean>(expressionData, subResolved, getType(Object), undefined) : false;

            if (subResolved.resolved && value != null && (!emptyAsNull || !(typeof value == 'string') || !String.isNullOrEmpty(value as string))) {
                return value;
            }
            else {
                return parameters.get(1).resolve(expressionData, subResolved, /* running:*/ undefined);
            }

        }

        return undefined;
    }
}
interface LooseObject {
    [key: string]: any
}
export class toObject extends mfunction {

    public /* override */ get minParameters(): number {
        return 0;
    }

    public /* override */ get maxParameters(): number {
        return -1;
    }

    public /* internal */ /* override */ invoke(expressionData: expressionData, parameters: List<expressionPart>, /* out */resolved: outputVariables, running: groupDataContext | undefined): Object | undefined {
        resolved.resolved = true;
        let source: Object | undefined = undefined;
        if ((parameters.count == 1)) {
            let subResolved: outputVariables = new outputVariables();
            source = parameters.get(0).resolve(expressionData, /* out */subResolved, /* running:*/ undefined);
        }
        else {
            source = expressionData.contexts.peek();
        }

        if ((source != null)) {
            if (source instanceof namedSelection || source instanceof DictionaryObject || source instanceof selection) {
                let eo: LooseObject = {};
                let eoColl = new List<KeyValuePair<string, Object>>();
                let dict = new DictionaryObject();
                if (source instanceof namedSelection) {
                    for (let strKey of source.getKeys()) {
                        dict.add(strKey.toString(), source.getValue(strKey) as object)
                    }
                }
                else if (source instanceof selection) {
                    let num = 0;
                    for (let strKey of source.toArray()) {
                        dict.add(num.toString(),strKey as object );
                        num++;
                    }
                }
                else {
                    dict = source as DictionaryObject;
                }

                for (let kvp of dict.getKeys()) {
                    //eoColl.add({ key: kvp, value: dict.getValue(kvp) as Object });
                    eo[kvp] = dict.getValue(kvp);
                }

                return (<LooseObject>(eo));
            }
            else {
                return source;
            }

        }

        return undefined;
    }
}

export class toObjectList extends mfunction {

    public /* override */ get minParameters(): number {
        return 0;
    }

    public /* override */ get maxParameters(): number {
        return -1;
    }

    public invoke(expressionData: expressionData, parameters: List<expressionPart>, /* out */resolved: outputVariables, running: groupDataContext | undefined): Object | undefined {
        resolved.resolved = true;
        let source: Object | undefined = undefined;
        let func: toObject = new toObject();
        if ((parameters.count == 1)) {
            let subResolved: outputVariables = new outputVariables();
            source = parameters.get(0).resolve(expressionData, /* out */subResolved, /* running:*/ undefined);
        }
        else {
            source = expressionData.contexts.peek();
        }

        if ((source != null)) {
            if ((source as IEnumerable<any>).getEnumerator() !== undefined) {
                let newList: List<Object> = new List<Object>();
                for (let item of (<List<any>>(source)).toArray()) {
                    expressionData.contexts.push(item);
                    let subResolved: outputVariables = new outputVariables();
                    newList.add(func.invoke(expressionData, new List<expressionPart>(), /* out */subResolved, running) as Object);
                    expressionData.contexts.pop();
                }

                return newList;
            }

            return source;
        }

        return undefined;
    }
}